from crystal_toolkit.components.transformations.core import TransformationComponent

from pymatgen.transformations.advanced_transformations import (
    MonteCarloRattleTransformation,
)


class MonteCarloRattleTransformationComponent(TransformationComponent):
    @property
    def title(self):
        return "Rattle a supercell"

    @property
    def description(self):
        return """Uses a Monte Carlo rattle procedure to randomly perturb the sites in a
    structure using the [hiPhive](https://hiphive.materialsmodeling.org) code.
    
Rattling atom \` i \` is carried out as a Monte Carlo move that is accepted with
a probability determined from the minimum interatomic distance
\` d\_{ij} \`.  If \` \\\\min(d\_{ij}) \` is smaller than \` d\_{min} \`
the move is only accepted with a low probability.

This process is repeated for each atom a number of times meaning
the magnitude of the final displacements is not *directly*
connected to the rattle amplitude.
"""

    @property
    def transformation(self):
        return MonteCarloRattleTransformation

    def options_layouts(self, state=None, structure=None):

        state = state or {
            "rattle_std": 0.2,
            "min_distance": 0.1,
            "seed": None,
        }

        rattle_std = self.get_numerical_input(
            label="Rattle amplitude",
            kwarg_label="rattle_std",
            state=state,
            help_str="""Rattle amplitude (standard deviation in normal distribution). 
Note: this value is not *directly* connected to the
final average displacement for the structures""",
            shape=(),
        )

        min_distance = self.get_numerical_input(
            label="Minimum distance /Å",
            kwarg_label="min_distance",
            state=state,
            help_str="""Interatomic distance used for computing the probability
for each rattle move.""",
            shape=(),
        )

        seed = self.get_numerical_input(
            label="Random seed",
            kwarg_label="seed",
            state=state,
            help_str="""Seed for setting up NumPy random state from which random numbers
are generated. If not set, a random seed will be generated
(default). This option allows the output of this transformation
to be deterministic.""",
            shape=(),
            is_int=True,
        )

        return [rattle_std, min_distance, seed]



from dash import dcc
from dash import html
from dash import dash_table as dt

from dash.dependencies import Input, Output, State

from crystal_toolkit.helpers.layouts import Label
from crystal_toolkit.components.transformations.core import TransformationComponent

from pymatgen.core.periodic_table import Specie, Element
from pymatgen.core.structure import Structure
from pymatgen.transformations.standard_transformations import SubstitutionTransformation

from ast import literal_eval


class SubstitutionTransformationComponent(TransformationComponent):
    @property
    def title(self):
        return "Substitute one species for another"

    @property
    def description(self):
        return """Replace one species in your structure (\"Previous Species\")
with another species (\"New Species\"). The new species can be specified as an
element (for example, O), as an element with an oxidation state (for example, O2-)
or as a composition (for example, {"Au":0.5, "Cu":0.5} for a 50/50 mixture of gold
and copper). Please consult the pymatgen documentation for more information.
"""

    @property
    def transformation(self):
        return SubstitutionTransformation

    def options_layouts(self, state=None, structure=None):

        if structure and structure.is_ordered:
            species_mapping = {
                el: None for el in [str(el) for el in structure.types_of_specie]
            }
        else:
            species_mapping = None

        state = state or {"species_mapping": species_mapping}

        species_mapping = self.get_dict_input(
            label="Species Mapping",
            kwarg_label="species_mapping",
            state=state,
            help_str="A mapping from an original species (element or element with oxidation state, e.g. O or O2-) "
            "to a new species (element, element with oxidation state, or a composition, e.g. O or O2- or "
            '{"Au": 0.5, "Cu": 0.5}). In pymatgen, these are Element, Specie and Composition classes '
            "respectively.",
            key_name="Original Species",
            value_name="New Species",
        )

        return [species_mapping]

    def generate_callbacks(self, app, cache):
        super().generate_callbacks(app, cache)
        #
        # @app.callback(
        #     Output(self.id("transformation_args_kwargs"), "data"),
        #     [Input(self.id("species_mapping"), "data")],
        # )
        # def update_transformation_kwargs(rows):
        #     def get_el_occu(string):
        #         try:
        #             el_occu = literal_eval(string)
        #         except ValueError:
        #             el_occu = string
        #         return el_occu
        #
        #     species_map = {
        #         get_el_occu(row["prev"]): get_el_occu(row["new"])
        #         for row in rows
        #         if (row["prev"] and row["new"])
        #     }
        #
        #     print(species_map)
        #
        #     return {"args": [species_map], "kwargs": {}}



from crystal_toolkit import _DEFAULTS

def update_object_args(d_args, object_name, allowed_args):
    """Read dafault properties and overwrite them if user input exists
    
    Arguments:
        d_args {dict} -- User defined properties
        object_name {str} -- Name of object
        allowed_kwargs {List[str]} -- Used to limit the data that is passed to pythreejs
    
    Returns:
        Dictionary -- Properties of object after userinput and default values are considered
    """
    obj_args = {
        k: v
        for k, v in (_DEFAULTS['scene'][object_name] or {}).items()
    }
    obj_args.update({
        k: v
        for k, v in (d_args or {}).items() if k in allowed_args and v != None
    })
    return obj_args


# standard Dash imports
import dash
from dash import dcc
from dash import html

# import for this example
from pymatgen.ext.matproj import MPRester
from pymatgen.analysis.phase_diagram import PhaseDiagram

# standard Crystal Toolkit import
import crystal_toolkit.components as ctc

# create Dash app as normal
app = dash.Dash()

# If callbacks created dynamically they cannot be statically checked at app startup.
# For this simple example this is not a problem, but if creating a complicated,
# nested layout this will need to be enabled -- consult Dash documentation
# for more information.
app.config["suppress_callback_exceptions"] = True

# tell Crystal Toolkit about the app
ctc.register_app(app)

# first, retrieve entries from Materials Project
with MPRester() as mpr:
    # li_entries = mpr.get_entries_in_chemsys(["Li"])
    # li_o_entries = mpr.get_entries_in_chemsys(["Li", "O"])
    li_co_o_entries = mpr.get_entries_in_chemsys(["Li", "O", "Co"])
    # li_co_o_fe_entries = mpr.get_entries_in_chemsys(["Li", "O", "Co", "Fe"])

# and then create the phase diagrams
# li_phase_diagram = PhaseDiagram(li_entries)
# li_o_phase_diagram = PhaseDiagram(li_o_entries)
li_co_o_phase_diagram = PhaseDiagram(li_co_o_entries)
# li_co_o_fe_phase_diagram = PhaseDiagram(li_co_o_fe_entries)

# and the corresponding Crystal Toolkit components
# we're creating four components here to demonstrate
# visualizing 1-D, 2-D, 3-D and 4-D phase diagrams
# li_phase_diagram_component = ctc.PhaseDiagramComponent(li_phase_diagram)
# li_o_phase_diagram_component = ctc.PhaseDiagramComponent(li_o_phase_diagram)
li_co_o_phase_diagram_component = ctc.PhaseDiagramComponent(li_co_o_phase_diagram)
# li_co_o_fe_phase_diagram_component = ctc.PhaseDiagramComponent(li_co_o_fe_phase_diagram)

print(li_co_o_entries)

# example layout to demonstrate capabilities of component
my_layout = html.Div(
    [
        html.H1("PhaseDiagramComponent Example"),
        html.H2("Standard Layout (1 Element)"),
        html.H2("Standard Layout (2 Elements)"),
        html.H2("Standard Layout (3 Elements)"),
        li_co_o_phase_diagram_component.layout(),
        html.H2("Standard Layout (4 Elements)"),
        html.H2("Technical Details"),
        dcc.Markdown(str(li_co_o_phase_diagram_component)),
    ]
)

# wrap your app.layout with crystal_toolkit_layout()
# to ensure all necessary components are loaded into layout
app.layout = ctc.crystal_toolkit_layout(my_layout)


# allow app to be run using "python structure.py"
# in production, deploy behind gunicorn or similar
# see Dash documentation for more information
if __name__ == "__main__":
    app.run_server(debug=True, port=8050)



# standard Dash imports
import dash
from dash import html
from dash import dcc

# standard Crystal Toolkit import
import crystal_toolkit.components as ctc
from crystal_toolkit.settings import SETTINGS
from crystal_toolkit.helpers.layouts import H1, H2, Container

# import for this example
from pymatgen.core.structure import Structure
from pymatgen.core.lattice import Lattice
from pymatgen.ext.matproj import MPRester
from pymatgen.analysis.phase_diagram import PhaseDiagram
from pymatgen.analysis.diffraction.xrd import XRDCalculator

# create Dash app as normal
app = dash.Dash(assets_folder=SETTINGS.ASSETS_PATH)

# create our crystal structure using pymatgen
from pymatgen.core.structure import Structure
from pymatgen.core.lattice import Lattice

structure = Structure(Lattice.cubic(4.2), ["Na", "K"], [[0, 0, 0], [0.5, 0.5, 0.5]])

xrd_component = ctc.XRayDiffractionComponent()

# example layout to demonstrate capabilities of component
my_layout = Container(
    [H1("XRDComponent Example (Empty, No Structure Defined)"), xrd_component.layout(),]
)

# as explained in "preamble" section in documentation
ctc.register_crystal_toolkit(app=app, layout=my_layout)

# allow app to be run using "python structure.py"
# in production, deploy behind gunicorn or similar
# see Dash documentation for more information
if __name__ == "__main__":
    app.run_server(debug=True, port=8050)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['crystal_toolkit',
 'crystal_toolkit.apps',
 'crystal_toolkit.apps.examples',
 'crystal_toolkit.apps.examples.tests',
 'crystal_toolkit.apps.tests',
 'crystal_toolkit.components',
 'crystal_toolkit.components.transformations',
 'crystal_toolkit.core',
 'crystal_toolkit.core.tests',
 'crystal_toolkit.helpers',
 'crystal_toolkit.renderables']

package_data = \
{'': ['*'], 'crystal_toolkit.apps': ['assets/*', 'assets/fonts/*']}

install_requires = \
['crystaltoolkit-extension>=0.4.0,<0.5.0',
 'mp-api',
 'mp-pyrho>=0.0.21,<0.0.22',
 'plotly>=5.3.1,<6.0.0',
 'pydantic',
 'pymatgen>=2022.0.14,<2023.0.0',
 'scikit-image',
 'scikit-learn',
 'webcolors']

extras_require = \
{'fermi': ['ifermi', 'pyfftw'],
 'figures': ['kaleido>=0.2.1,<0.3.0'],
 'server': ['dash>=2.0.0,<3.0.0',
            'dash-daq',
            'gunicorn',
            'redis',
            'Flask-Caching',
            'gevent',
            'dash-mp-components>=0.3.44,<0.4.0',
            'robocrys',
            'habanero',
            'dscribe',
            'dash-extensions',
            'sentry-sdk'],
 'vtk': ['dash-vtk>=0.0.6,<0.0.7']}

setup_kwargs = {
    'name': 'crystal-toolkit',
    'version': '2021.10.21',
    'description': '',
    'long_description': None,
    'author': 'Matthew Horton',
    'author_email': 'mkhorton@lbl.gov',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import logging
from abc import ABC, abstractmethod
from ast import literal_eval
from collections import defaultdict
from itertools import chain
from json import JSONDecodeError, dumps, loads
from typing import Dict, List, Optional, Set, Tuple, Union
from base64 import b64encode

import dash
from dash import dcc
from dash import html
from dash import dash_table as dt
import numpy as np
from dash.dependencies import ALL, Output
from flask_caching import Cache
from monty.json import MontyDecoder, MSONable

from crystal_toolkit import __version__ as ct_version
from crystal_toolkit.helpers.layouts import add_label_help, Icon, Button, Loading
from crystal_toolkit.settings import SETTINGS

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

# fallback cache if Redis etc. isn't set up
null_cache = Cache(config={"CACHE_TYPE": "null"})

# Crystal Toolkit namespace, added to the start of all ids
# so we can see which layouts have been added by Crystal Toolkit
CT_NAMESPACE = "CT"


class MPComponent(ABC):
    """
    The abstract base class for an MPComponent. MPComponent
    is designed to help render an MSONable object.
    """

    # reference to global Dash app
    app = None

    # reference to Flask cache
    cache = None

    # used to track all dcc.Stores required for all MPComponents to work
    # keyed by the MPComponent id
    _app_stores_dict: Dict[str, List[dcc.Store]] = defaultdict(list)

    # used to track what individual Dash components are defined
    # by this MPComponent
    _all_id_basenames: Set[str] = set()

    # used to defer generation of callbacks until app.layout defined
    # can be helpful to callback exceptions retained
    _callbacks_to_generate: Set["MPComponent"] = set()

    @staticmethod
    def register_app(app: dash.Dash):
        """
        This method must be called at least once in your Crystal
        Toolkit Dash app if you want to enable interactivity with the
        MPComponents. The "app" variable is a special global
        variable used by Dash/Flask, and registering it with
        MPComponent allows callbacks to be registered with the
        app on instantiation.

        Args:
            app: a Dash app instance
        """
        MPComponent.app = app
        # add metadata
        app.config.meta_tags.append(
            {
                "name": "generator",
                "content": f"Crystal Toolkit {ct_version} (Materials Project)",
            }
        )
        # set default title, but respect the user if they override it
        if app.title == "Dash":
            app.title = "Crystal Toolkit"

    @staticmethod
    def register_cache(cache: Cache):
        """
        This method must be called at least once in your
        Crystal Toolkit Dash app if you want to enable
        callback caching. Callback caching is one of the
        easiest ways to see significant performance
        improvements, especially for callbacks that are
        computationally expensive.

        Args:
            cache: a flask_caching Cache instance
        """
        if cache:
            MPComponent.cache = cache
        else:
            MPComponent.cache = Cache(
                MPComponent.app.server, config={"CACHE_TYPE": "simple"}
            )

    @staticmethod
    def crystal_toolkit_layout(layout: html.Div) -> html.Div:

        if not MPComponent.app:
            raise ValueError(
                "Please register the Dash app with Crystal Toolkit "
                "using register_app()."
            )

        # layout_str = str(layout)
        stores_to_add = []
        for basename in MPComponent._all_id_basenames:
            # can use "if basename in layout_str:" to restrict to components present in initial layout
            # this would cause bugs for components displayed dynamically
            stores_to_add += MPComponent._app_stores_dict[basename]
        layout.children += stores_to_add

        # set app.layout to layout so that callbacks can be validated
        MPComponent.app.layout = layout

        for component in MPComponent._callbacks_to_generate:
            component.generate_callbacks(MPComponent.app, MPComponent.cache)

        return layout

    @staticmethod
    def register_crystal_toolkit(app, layout, cache=None):

        MPComponent.register_app(app)
        MPComponent.register_cache(cache)
        app.config["suppress_callback_exceptions"] = True
        app.layout = MPComponent.crystal_toolkit_layout(layout)

    @staticmethod
    def all_app_stores() -> html.Div:
        """
        This must be included somewhere in your
        Crystal Toolkit Dash app's layout for
        interactivity to work. This is a hidden element
        that contains the MSON for each MPComponent.

        Returns: a html.Div Dash Layout
        """
        return html.Div(
            list(chain.from_iterable(MPComponent._app_stores_dict.values()))
        )

    def __init__(
        self,
        default_data: Optional[Union[MSONable, Dict, str]] = None,
        id: Optional[str] = None,
        links: Optional[Dict[str, str]] = None,
        storage_type: Literal["memory", "local", "session"] = "memory",
        disable_callbacks: bool = False,
    ):
        """
        The abstract base class for an MPComponent.

        The MPComponent is designed to help render any MSONable object,
        for example many of the objects in pymatgen (Structure, PhaseDiagram, etc.)

        To instantiate an MPComponent, you will need to create it outside
        of your Dash app layout:

        my_component = MPComponent(my_msonable_object)

        Then, inside the app.layout, you can include the component's layout
        anywhere you choose: my_component.layout

        If you want the layouts to be interactive, i.e. to respond to callbacks,
        you have to also use the MPComponent.register_app(app) method in your app,
        and also include MPComponent.all_app_stores in your app.layout (an
        invisible layout that contains the MSON itself).

        If you do not want the layouts to be interactive, set disable_callbacks
        to True to prevent errors.

        If including multiple MPComponents of the same type, make sure
        to set the id field to a unique value, as you would in any other
        Dash component.

        When sub-classing MPComponent, the most important methods to implement
        are _sub_layouts and generate_callbacks().

        Args:
            default_data: inital contents for the component, can be None
            id: a unique id, required if multiple of the same type of
            MPComponent are included in an app
            links: if set, will set store contents from the stores of another
            component to reduce unnecessary callbacks and duplication of data,
            note that links are one directional only and specific the origin
            stores, e.g. set {"default": my_other_component.id()} to fill this
            component's default store contents from the other component's default store,
            or {"graph": my_other_component.id("graph")} to fill this component's
            "graph" store from another component's "graph" store
            storage_type: whether to persist contents of component through
            browser refresh or browser sessions, use with caution, defaults
            to "memory" so component store will be emptied on refresh, see
            dcc.Store documentation for more information
            disable_callbacks: if True, will not generate callbacks, useful
            for static layouts or returning new MPComponents dynamically where
            generating callbacks are not possible due to limitations of Dash
        """

        # ensure ids are unique
        # Note: shadowing Python built-in here, but only because Dash does it...
        # TODO: do something else here
        if id is None:
            # TODO: this could lead to duplicate ids and an error, but if
            # setting random ids, this could also lead to undefined behavior
            id = f"{CT_NAMESPACE}{self.__class__.__name__}"
        elif not id.startswith(CT_NAMESPACE):
            id = f"{CT_NAMESPACE}{id}"
        MPComponent._all_id_basenames.add(id)

        self._id = id
        self._all_ids = set()
        self._stores = {}
        self._initial_data = {}

        self.links = links or {}

        self.create_store(
            name="default", initial_data=default_data, storage_type=storage_type
        )
        self.links["default"] = self.id()

        if not disable_callbacks:
            # callbacks generated as final step by crystal_toolkit_layout()
            self._callbacks_to_generate.add(self)

        self.logger = logging.getLogger(self.__class__.__name__)

    def id(
        self,
        name: str = "default",
        is_kwarg: bool = False,
        idx=False,
        hint=None,
        is_store: bool = False,
    ) -> Union[str, Dict[str, str]]:
        """
        Generate an id from a name combined with the
        base id of the MPComponent itself, useful for generating
        ids of individual components in the layout.

        In the special case of the id of an element that is used to re-construct
        a keyword argument for a specific class, it will store information necessary
        to reconstruct that keyword argument (e.g. its type hint and, in the case of
        a vector or matrix, the corresponding index).

        A hint could be a tuple for a numpy array of that shape, e.g. (3, 3) for a 3x3 matrix,
        (1, 3) for a vector, or "literal" to parse kwarg value using ast.literal_eval, or "bool"
        to parse a boolean value. In future iterations, we may be able to replace this with native
        Python type hints. The problem here is being able to specify array shape where appropriate.


        Args:
            name: e.g. "default"

        Returns: e.g. "MPComponent_default"
        """

        if name in self._stores:
            is_store = True

        if is_kwarg:
            return {
                "component_id": self._id,
                "kwarg_label": name,
                "idx": str(idx),
                "hint": str(hint),
            }

        # if we're linking to another component, return that id
        if name in self.links:
            return self.links[name]

        # otherwise create a new id
        self._all_ids.add(name)
        if name != "default":
            name = f"{self._id}_{name}"
        else:
            name = f"{self._id}"
        return name
        if is_store:
            return name
        else:
            return {"id": name}

    def create_store(
        self,
        name: str,
        initial_data: Optional[Union[MSONable, Dict, str]] = None,
        storage_type: Literal["memory", "local", "session"] = "memory",
        debug_clear: bool = False,
    ):
        """
        Generate a dcc.Store to hold something (MSONable object, Dict
        or string), and register it so that it will be included in the
        Dash app automatically.

        The initial data will be stored in a class attribute as
        self._initial_data[name].

        Args:
            name: name for the store
            initial_data: initial data to include
            storage_type: as in dcc.Store
            debug_clear: set to True to empty the store if using
            persistent storage
        """

        # if we're linking to another component, do not create a new store
        if name in self.links:
            return

        store = dcc.Store(
            id=self.id(name, is_store=True),
            data=initial_data,
            storage_type=storage_type,
            clear_data=debug_clear,
        )
        self._stores[name] = store
        self._initial_data[name] = initial_data
        MPComponent._app_stores_dict[self.id()].append(store)

    @property
    def initial_data(self):
        """
        :return: Initial data for all the stores defined by component,
        keyed by store name.
        """
        return self._initial_data

    @staticmethod
    def from_data(data):
        """
        Converts the contents of a dcc.Store back into a Python object.
        :param data: contents of a dcc.Store created by to_data
        :return: a Python object
        """
        return loads(dumps(data), cls=MontyDecoder)

    @property
    def all_stores(self) -> List[str]:
        """
        :return: List of all store ids generated by this component
        """
        return list(self._stores.keys())

    @property
    def all_ids(self) -> List[str]:
        """
        :return: List of all ids generated by this component
        """
        return list(
            [
                component_id
                for component_id in self._all_ids
                if component_id not in self.all_stores
            ]
        )

    def __repr__(self):
        return f"{self.id()}<{self.__class__.__name__}>"

    def __str__(self):
        ids = "\n".join(
            [f"* {component_id}  " for component_id in sorted(self.all_ids)]
        )
        stores = "\n".join([f"* {store}  " for store in sorted(self.all_stores)])
        layouts = "\n".join(
            [f"* {layout}  " for layout in sorted(self._sub_layouts.keys())]
        )

        return f"""{self.id()}<{self.__class__.__name__}>  \n
IDs:  \n{ids}  \n
Stores:  \n{stores}  \n
Sub-layouts:  \n{layouts}"""

    @property
    def _sub_layouts(self):
        """
        Layouts associated with this component, available for book-keeping
        if your component is complex, so that the layout() method is just
        assembles individual sub-layouts.

        :return: A dictionary with names of layouts as keys (str) and Dash
        layouts (e.g. html.Div) as values.
        """
        return {}

    def layout(self) -> html.Div:
        """
        :return: A Dash layout for the full component. Basic implementation
        provided, but should in general be overridden.
        """
        return html.Div(list(self._sub_layouts.values()))

    @abstractmethod
    def generate_callbacks(self, app, cache):
        """
        Generate all callbacks associated with the layouts in this app. Assume
        that "suppress_callback_exceptions" is True, since it is not always
        guaranteed that all layouts will be displayed to the end user at all
        times, but it's important the callbacks are defined on the server.
        """
        raise NotImplementedError

    def get_numerical_input(
        self,
        kwarg_label: str,
        state: Optional[dict] = None,
        label: Optional[str] = None,
        help_str: str = None,
        is_int: bool = False,
        shape: Tuple[int, ...] = (),
        **kwargs,
    ):
        """
        For Python classes which take matrices as inputs, this will generate
        a corresponding Dash input layout.

        :param kwarg_label: The name of the corresponding Python input, this is used
        to name the component.
        :param label: A description for this input.
        :param state: Used to set state for this input, dict with arg name or kwarg name as key
        :param help_str: Text for a tooltip when hovering over label.
        :param is_int: if True, will use a numeric input
        :param shape: (3, 3) for matrix, (1, 3) for vector, (1, 1) for scalar
        :return: a Dash layout
        """

        default = np.full(shape, state.get(kwarg_label))
        default = np.reshape(default, shape)

        def matrix_element(idx, value=0):
            # TODO: maybe move element out of the name
            mid = self.id(kwarg_label, is_kwarg=True, idx=idx, hint=shape)
            if not is_int:
                return dcc.Input(
                    id=mid,
                    inputMode="numeric",
                    debounce=True,
                    className="input",
                    style={
                        "textAlign": "center",
                        # shorter default width if matrix or vector
                        "width": "5rem",
                        "marginRight": "0.2rem",
                        "marginBottom": "0.2rem",
                        "height": "36px",
                    },
                    value=value,
                    persistence=True,
                    type="number",
                    **kwargs,
                )
            else:
                return dcc.Input(
                    id=mid,
                    inputMode="numeric",
                    debounce=True,
                    className="input",
                    style={
                        "textAlign": "center",
                        # shorter default width if matrix or vector
                        "width": "5rem",
                        "marginRight": "0.2rem",
                        "marginBottom": "0.2rem",
                        "height": "36px",
                    },
                    value=value,
                    persistence=True,
                    type="number",
                    step=1,
                    **kwargs,
                )

        # dict of row indices, column indices to element
        matrix_contents = defaultdict(dict)

        # determine what individual input boxes we need
        # note that shape = () for floats, shape = (3,) for vectors
        # but we may also need to accept input for e.g. (3, 1)
        it = np.nditer(default, flags=["multi_index", "refs_ok"])
        while not it.finished:
            idx = it.multi_index
            row = (idx[1] if len(idx) > 1 else 0,)
            column = idx[0] if len(idx) > 0 else 0
            matrix_contents[row][column] = matrix_element(idx, value=it[0])
            it.iternext()

        # arrange the input boxes in two dimensions (rows, columns)
        matrix_div_contents = []
        for row_idx, columns in sorted(matrix_contents.items()):
            row = []
            for column_idx, element in sorted(columns.items()):
                row.append(element)
            matrix_div_contents.append(html.Div(row))

        matrix = html.Div(matrix_div_contents)

        return add_label_help(matrix, label, help_str)

    def get_slider_input(
        self,
        kwarg_label: str,
        state: Optional[dict] = None,
        label: Optional[str] = None,
        help_str: str = None,
        multiple: bool = False,
        **kwargs,
    ):

        default = state.get(kwarg_label) or False

        if multiple:
            slider_input = dcc.RangeSlider(
                id=self.id(kwarg_label, is_kwarg=True, hint="slider"),
                tooltip={"placement": "bottom"},
                value=default,
                persistence=True,
                **kwargs,
            )
        else:
            slider_input = dcc.Slider(
                id=self.id(kwarg_label, is_kwarg=True, hint="slider"),
                tooltip={"placement": "bottom"},
                value=default,
                persistence=True,
                **kwargs,
            )

        return add_label_help(slider_input, label, help_str)

    def get_bool_input(
        self,
        kwarg_label: str,
        state: Optional[dict] = None,
        label: Optional[str] = None,
        help_str: str = None,
    ):
        """
        For Python classes which take boolean values as inputs, this will generate
        a corresponding Dash input layout.

        :param kwarg_label: The name of the corresponding Python input, this is used
        to name the component.
        :param label: A description for this input.
        :param state: Used to set state for this input, dict with arg name or kwarg name as key
        :param help_str: Text for a tooltip when hovering over label.
        :return: a Dash layout
        """

        default = state.get(kwarg_label) or False

        bool_input = dcc.Checklist(
            id=self.id(kwarg_label, is_kwarg=True, hint="bool"),
            style={"width": "5rem"},
            options=[{"label": "", "value": "enabled"}],
            value=["enabled"] if default else [],
            persistence=True,
        )

        return add_label_help(bool_input, label, help_str)

    def get_choice_input(
        self,
        kwarg_label: str,
        state: Optional[dict] = None,
        label: Optional[str] = None,
        help_str: str = None,
        options: Optional[List[Dict]] = None,
        **kwargs,
    ):
        """
        For Python classes which take floats as inputs, this will generate
        a corresponding Dash input layout.

        :param kwarg_label: The name of the corresponding Python input, this is used
        to name the component.
        :param label: A description for this input.
        :param state: Used to set state for this input, dict with arg name or kwarg name as key
        :param help_str: Text for a tooltip when hovering over label.
        :param options: Options to choose from, as per dcc.Dropdown
        :return: a Dash layout
        """

        default = state.get(kwarg_label)

        option_input = dcc.Dropdown(
            id=self.id(kwarg_label, is_kwarg=True, hint="literal"),
            options=options if options else [],
            value=default,
            persistence=True,
            clearable=False,
            **kwargs,
        )

        return add_label_help(option_input, label, help_str)

    def get_dict_input(
        self,
        kwarg_label: str,
        state: Optional[dict] = None,
        label: Optional[str] = None,
        help_str: str = None,
        key_name: str = "key",
        value_name: str = "value",
    ):

        default = state.get(kwarg_label) or {}

        dict_input = dt.DataTable(
            id=self.id(kwarg_label, is_kwarg=True, hint="dict"),
            columns=[
                {"id": "key", "name": key_name},
                {"id": "value", "name": value_name},
            ],
            data=[{"key": k, "value": v} for k, v in default.items()],
            editable=True,
            persistence=True,
        )

        return add_label_help(dict_input, label, help_str)

    def get_kwarg_id(self, kwarg_name) -> Dict:
        """

        :param kwarg_name:
        :return:
        """
        return {
            "component_id": self._id,
            "kwarg_label": kwarg_name,
            "idx": ALL,
            "hint": ALL,
        }

    def get_all_kwargs_id(self) -> Dict:
        """

        :return:
        """
        return {"component_id": self._id, "kwarg_label": ALL, "idx": ALL, "hint": ALL}

    def reconstruct_kwarg_from_state(self, state, kwarg_name):
        return self.reconstruct_kwargs_from_state(state)[kwarg_name]

    def reconstruct_kwargs_from_state(self, state) -> Dict:
        # TODO: change to take callback_context directly

        kwargs = {}
        for k, v in state.items():

            # TODO: hopefully this will be less hacky in future Dash versions
            # remove trailing ".value" and convert back into dictionary
            # need to sort k somehow ...

            try:
                d = loads(k[: -len(".value")])
            except JSONDecodeError:
                continue

            kwarg_label = d["kwarg_label"]

            try:
                k_type = literal_eval(d["hint"])
            except ValueError:
                k_type = d["hint"]

            idx = literal_eval(d["idx"])

            # TODO: catch Exceptions here, and display validation error to user if incorrect kwargs supplied

            if isinstance(k_type, tuple):
                # matrix or vector
                if kwarg_label not in kwargs:
                    kwargs[kwarg_label] = np.empty(k_type)
                v = literal_eval(str(v))
                if v is not None and kwargs[kwarg_label] is not None:
                    kwargs[kwarg_label][idx] = literal_eval(str(v))
                else:
                    # require all elements to have value, otherwise set
                    # entire kwarg to None
                    kwargs[kwarg_label] = None

            elif k_type == "literal":

                try:
                    kwargs[kwarg_label] = literal_eval(str(v))
                except ValueError:
                    kwargs[kwarg_label] = str(v)

            elif k_type == "bool":
                kwargs[kwarg_label] = bool("enabled" in v)

            elif k_type == "slider":
                kwargs[kwarg_label] = v

            elif k_type == "dict":
                pass

        for k, v in kwargs.items():
            if isinstance(v, np.ndarray):
                kwargs[k] = v.tolist()

        if SETTINGS.DEBUG_MODE:
            print(self.__class__.__name__, "kwargs", kwargs)

        return kwargs

    @staticmethod
    def datauri_from_fig(
        fig, fmt: str = "png", width: int = 600, height: int = 400, scale: int = 4
    ) -> str:
        """
        Generate a data URI from a Plotly Figure.

        :param fig: Plotly Figure object or corresponding dictionary
        :param fmt: "png", "jpg", etc. (see PlotlyScope for supported formats)
        :param width: width in pixels
        :param height: height in pixels
        :param scale: scale factor
        :return:
        """

        from kaleido.scopes.plotly import PlotlyScope

        scope = PlotlyScope()
        output = scope.transform(
            fig, format=fmt, width=width, height=height, scale=scale
        )
        image = b64encode(output).decode("ascii")

        return f"data:image/{fmt};base64,{image}"

    def get_figure_placeholder(self, figure_id: str) -> html.Div:
        """
        Get a layout to act as a placeholder for an interactive figure.

        When used with `generate_static_figure_callbacks`, and assuming
        kaleido is installed on the server, a static image placeholder will
        be generated.

        :return:
        """

        return html.Div(
            [
                html.Div(
                    [Loading(id=self.id(f"{figure_id}-wrapped-figure-inner"))],
                    id=self.id("wrapped-figure-outer"),
                ),
                Button(
                    [Icon(kind="chart-pie"), html.Span(), "Make Plot Interactive"],
                    kind="primary",
                    id=self.id(f"{figure_id}-wrapped-figure-button"),
                ),
            ]
        )



from crystal_toolkit.apps.examples.bandstructure import app
import time


def test_bs(dash_duo):

    dash_duo.start_server(app)
    dash_duo.clear_storage()

    time.sleep(5)
    dash_duo.percy_snapshot("example_bsdos_on_load")

    # # test choosing elemental projection
    # el = dash_duo.select_dcc_dropdown(
    #     '{"component_id":"CTbs_dos","hint":"literal","idx":"False","kwarg_label":"dos-select"}',
    #     index=1,
    # )
    #
    # time.sleep(3)
    # dash_duo.percy_snapshot("example_bsdos_projection_index_1")
    #
    # # test selecting total orbital projection
    # el = dash_duo.select_dcc_dropdown("#CTbs_dos_dos-select", index=2)
    #
    # time.sleep(3)
    # dash_duo.percy_snapshot("example_bsdos_projection_index_2")

    assert (
        bool(dash_duo.get_logs()) is False
    ), f"Browser console contains an error: {dash_duo.get_logs()}"



import json
import os as _os
from collections import defaultdict
from pathlib import Path

# pleasant hack to support MSONable objects in Dash callbacks natively
from monty.json import MSONable

from crystal_toolkit.renderables import *

__version__ = "2021.10.21"

MODULE_PATH = Path(__file__).parents[0]


def to_plotly_json(self):
    return self.as_dict()


MSONable.to_plotly_json = to_plotly_json


# Populate the default values from the JSON file
_DEFAULTS = defaultdict(lambda: None)
default_js = _os.path.join(
    _os.path.join(_os.path.dirname(_os.path.abspath(__file__))), "./", "defaults.json"
)

with open(default_js) as handle:
    _DEFAULTS.update(json.loads(handle.read()))


def _repr_mimebundle_(self, include=None, exclude=None):
    """
    Render Scenes using crystaltoolkit-extension for Jupyter Lab.
    """
    # TODO: add Plotly, application/vnd.plotly.v1+json

    help_text_ct = """If you see this text, the Crystal Toolkit Jupyter Lab \n
extension is not installed. You can install it by running \n
\"pip install crystaltoolkit-extension\" \n
from the same environment you run \"jupyter lab\". \n
This only works in Jupyter Lab 3.x or above.\n\n
"""

    help_text_plotly = """If you see this text, the Plotly Jupyter Lab extension 
is not installed, please consult Plotly documentation for information on how to 
install.
"""

    # TODO: to be strict here, we could use inspect.signature
    # and .return_annotation is either a Scene or a go.Figure respectively
    # and also check all .parameters .kind.name have no POSITIONAL_ONLY
    # in practice, fairly unlikely this will cause issues without strict checking

    if hasattr(self, "get_scene"):
        return {
            "application/vnd.mp.ctk+json": self.get_scene().to_json(),
            "text/plain": help_text_ct + self.__repr__(),
        }
    elif hasattr(self, "get_plot"):
        return {
            "application/vnd.plotly.v1+json": self.get_plot().to_plotly_json(),
            "text/plain": help_text_plotly + self.__repr__(),
        }
    else:
        return {"application/json": self.as_dict(), "text/plain": self.__repr__()}


MSONable._repr_mimebundle_ = _repr_mimebundle_


def show_json(self):
    from IPython.display import display_json

    return display_json(self.as_dict(), raw=True)


MSONable.show_json = show_json


def _ipython_display_(self):
    """
    Render Scenes using crystaltoolkit-extension for Jupyter Lab.

    This function ensures that objects are also printed in string format
    as previously.
    """
    from IPython.display import publish_display_data

    publish_display_data(self._repr_mimebundle_())


MSONable._ipython_display_ = _ipython_display_



from dash import dcc
from dash import html

from dash.dependencies import Input, Output, State

from crystal_toolkit.helpers.layouts import Label
from crystal_toolkit.components.transformations.core import TransformationComponent
from dash_mp_components import CrystalToolkitScene

from crystal_toolkit.core.scene import Scene

from pymatgen.transformations.standard_transformations import SupercellTransformation


class SupercellTransformationComponent(TransformationComponent):
    @property
    def title(self):
        return "Make a supercell"

    @property
    def description(self):
        return """Create a supercell by providing a scaling matrix that transforms
input lattice vectors a, b and c into transformed lattice vectors a', b' and c'.


For example, to create a supercell with lattice vectors a'=2a, b'=2b, c'=2c enter a 
scaling matrix [[2, 0, 0], [0, 2, 0], [0, 0, 2]] or to create a supercell with 
lattice vectors a' = 2a+b, b' = 3b and c' = c enter a scaling matrix 
[[2, 1, 0], [0, 3, 0], [0, 0, 1]]. All elements of the scaling matrix must be 
integers."""

    @property
    def transformation(self):
        return SupercellTransformation

    def options_layouts(self, state=None, structure=None):

        state = state or {"scaling_matrix": ((1, 0, 0), (0, 1, 0), (0, 0, 1))}

        options = self.get_numerical_input(
            label="Scaling matrix",
            kwarg_label="scaling_matrix",
            state=state,
            help_str="""Scaling matrix that transforms 
            input lattice vectors a, b and c into transformed lattice vectors a', b' and c'.""",
            shape=(3, 3),
        )

        return [options]

    def get_preview_layout(self, struct_in, struct_out):

        if struct_in.lattice == struct_out.lattice:
            return html.Div()

        lattice_in = struct_in.lattice.get_scene()
        lattice_out = struct_out.lattice.get_scene(color="red")

        scene = Scene("lattices", contents=[lattice_in, lattice_out])

        return html.Div(
            [CrystalToolkitScene(data=scene.to_json())],
            style={"width": "100px", "height": "100px"},
        )



# standard Dash imports
import dash
from dash import dcc
from dash import html

# import for this example
from pymatgen.core.structure import Structure
from pymatgen.core.lattice import Lattice
from pymatgen.electronic_structure.core import Magmom

# standard Crystal Toolkit import
import crystal_toolkit.components as ctc

# create Dash app as normal
app = dash.Dash()

# create the Structure object
structure = Structure(
    Lattice.cubic(3.0),
    ["Ni", "Ti"],
    [[0, 0, 0], [0.5, 0.5, 0.5]],
    site_properties={"magmom": [[-2.0, 1.0, 0.0], [1.0, 1.0, -1.0]]},
    # site_properties={"magmom": [3.0, -2.0]},
)

# create the Crystal Toolkit component
structure_component = ctc.StructureMoleculeComponent(structure, id="struct")

# example layout to demonstrate capabilities of component
my_layout = html.Div(
    [
        html.H1("StructureMoleculeComponent Example"),
        html.H2("Standard Layout"),
        structure_component.layout(size="400px"),
        html.H2("Optional Additional Layouts"),
        html.H3("Screenshot Layout"),
        structure_component.screenshot_layout(),
        html.H3("Options Layout"),
        structure_component.options_layout(),
        html.H3("Title Layout"),
        structure_component.title_layout(),
        html.H3("Legend Layout"),
        structure_component.legend_layout(),
    ]
)

# tell crystal toolkit about your app and layout
ctc.register_crystal_toolkit(app, layout=my_layout)

# allow app to be run using "python structure.py"
# in production, deploy behind gunicorn or similar
# see Dash documentation for more information
if __name__ == "__main__":
    app.run_server(debug=True, port=8050)



from dash import dcc
from dash import html

from dash.dependencies import Input, Output, State

from crystal_toolkit.helpers.layouts import Label
from crystal_toolkit.components.transformations.core import TransformationComponent

from pymatgen.transformations.advanced_transformations import SlabTransformation


class SlabTransformationComponent(TransformationComponent):
    @property
    def title(self):
        return "Make a slab"

    @property
    def description(self):
        return """Create a slab from a structure, where a "slab" is a crystal 
surface that is still periodic in all three dimensions but has a large artificial
vacuum inserted so that the properties of the crystal surface can be studied.
"""

    @property
    def transformation(self):
        return SlabTransformation

    def options_layouts(self, state=None, structure=None):

        state = state or {
            "miller_index": (0, 0, 1),
            "min_slab_size": 4,
            "min_vacuum_size": 10,
            "lll_reduce": True,
            "center_slab": True,
            "in_unit_planes": False,
            "primitive": True,
            "max_normal_search": None,
            "shift": 0,
            "tol": 0.1,
        }

        miller_index = self.get_numerical_input(
            label="Miller index",
            kwarg_label="miller_index",
            state=state,
            help_str="The surface plane defined by its Miller index (h, k, l)",
            shape=(3,),
        )

        min_slab_size = self.get_numerical_input(
            label="Minimum slab size /Å",
            kwarg_label="min_slab_size",
            state=state,
            help_str="Minimum slab size in Ångstroms (or number of planes of atoms if "
            '"Use plane units" enabled)',
            shape=(3,),
        )

        min_vacuum_size = self.get_numerical_input(
            label="Minimum vacuum size /Å",
            kwarg_label="min_vacuum_size",
            state=state,
            help_str="Minimum vacuum size in Ångstroms (or number of planes of atoms if "
            '"Use plane units" enabled)',
            shape=(),
        )

        lll_reduce = self.get_bool_input(
            label="Enable LLL reduction",
            kwarg_label="lll_reduce",
            state=state,
            help_str="Whether or not to apply an LLL lattice reduction",
        )

        in_unit_planes = self.get_bool_input(
            label="Use Plane Units",
            kwarg_label="in_unit_planes",
            state=state,
            help_str="Change units of vacuum size and slab size to be in terms of "
            "number of planes of atoms instead of Ångstroms.",
        )

        primitive = self.get_bool_input(
            label="Make primitive",
            kwarg_label="primitive",
            state=state,
            help_str="Reduce the slab to most primitive cell.",
        )

        max_normal_search = self.get_numerical_input(
            "max_normal_search",
            state=state,
            label="Maximum normal search index",
            help_str="Maximum index to include in linear combinations of indices "
            "to find **c** lattice vector orthogonal to slab surface.",
            is_int=True,
        )

        shift = self.get_numerical_input(
            label="Shift /Å",
            kwarg_label="shift",
            state=state,
            help_str="Shift to change termination.",
            shape=(),
        )

        tol = self.get_numerical_input(
            label="Tolerance",
            kwarg_label="tol",
            state=state,
            help_str="Tolerance to find primitive cell.",
            shape=(),
        )

        options = html.Div(
            [
                miller_index,
                min_slab_size,
                min_vacuum_size,
                lll_reduce,
                in_unit_planes,
                primitive,
                max_normal_search,
                shift,
                tol,
            ]
        )

        return options



from collections import defaultdict
from itertools import combinations

import numpy as np
from pymatgen.analysis.graphs import MoleculeGraph

from crystal_toolkit.core.scene import Scene
from crystal_toolkit.core.legend import Legend


# TODO: fix Sam's bug (reorder)


def get_molecule_graph_scene(
    self,
    origin=None,
    explicitly_calculate_polyhedra_hull=False,
    legend=None,
    draw_polyhedra=False,
) -> Scene:

    legend = legend or Legend(self.molecule)

    primitives = defaultdict(list)

    for idx, site in enumerate(self.molecule):

        connected_sites = self.get_connected_sites(idx)

        site_scene = site.get_scene(
            connected_sites=connected_sites,
            origin=origin,
            explicitly_calculate_polyhedra_hull=explicitly_calculate_polyhedra_hull,
            legend=legend,
            draw_polyhedra=draw_polyhedra,
        )
        for scene in site_scene.contents:
            primitives[scene.name] += scene.contents

    return Scene(
        name=self.molecule.composition.reduced_formula,
        contents=[Scene(name=k, contents=v) for k, v in primitives.items()],
        origin=origin if origin else (0, 0, 0),
    )


MoleculeGraph.get_scene = get_molecule_graph_scene



from pymatgen.core.periodic_table import Specie, Element
from pymatgen.core.structure import Molecule
from pymatgen.core.structure import SiteCollection, Site
from pymatgen.analysis.molecule_structure_comparator import CovalentRadius
from pymatgen.util.string import unicodeify_species

from monty.json import MSONable
from monty.serialization import loadfn

from itertools import chain
from collections import defaultdict

from palettable.colorbrewer.qualitative import Set1_9, Set2_8
from sklearn.preprocessing import LabelEncoder
from matplotlib.cm import get_cmap
from webcolors import html5_parse_legacy_color, html5_serialize_simple_color

from typing import Union, Optional, Tuple, Dict, List, Any

import warnings
import numpy as np
import os

# element colors forked from pymatgen
module_dir = os.path.dirname(os.path.abspath(__file__))
EL_COLORS = loadfn(os.path.join(module_dir, "ElementColorSchemes.yaml"))


class Legend(MSONable):
    """
    Help generate a legend (colors and radii) for a Structure or Molecule
    such that colors and radii can be displayed for the appropriate species.

    Note that species themselves have a color (for example, Oxygen is typically
    red), but that we might also want to color-code by site properties (for example,
    magnetic moment), thus this class has to take into account both the species
    present and its context (the specific site the species is at) to correctly
    generate the legend.
    """

    default_color_scheme = "Jmol"
    default_color = [0, 0, 0]
    default_radius = 1.0
    fallback_radius = 0.5
    uniform_radius = 0.5

    def __init__(
        self,
        site_collection: Union[SiteCollection, Site],
        color_scheme: str = "Jmol",
        radius_scheme: str = "uniform",
        cmap: str = "coolwarm",
        cmap_range: Optional[Tuple[float, float]] = None,
    ):
        """
        Create a legend for a given SiteCollection to choose how to
        display colors and radii for the given sites and the species
        on those sites.

        If a site has a "display_color" or "display_radius" site
        property defined, this can be used to manually override the
        displayed colors and radii respectively.

        Args:
            site_collection: SiteCollection or, for convenience, a
            single site can be provided and this will be converted
            into a SiteCollection
            color_scheme: choose how to color-code species, one of
            "Jmol", "VESTA", "accessible" or a scalar site property
            (e.g. magnetic moment) or a categorical/string site
            property (e.g. Wyckoff label)
            radius_scheme: choose the radius for a species, one of
            "atomic", "specified_or_average_ionic", "covalent",
            "van_der_waals", "atomic_calculated", "uniform"
            cmap: only used if color_mode is set to a scalar site
            property, defines the matplotlib color map to use, by
            default is blue-white-red for negative to postive values
            cmap_range: only used if color_mode is set to a scalar site
            property, defines the minimum and maximum values of the
            color scape
        """

        if isinstance(site_collection, Site):
            site_collection = Molecule.from_sites([site_collection])

        site_prop_types = self.analyze_site_props(site_collection)

        self.allowed_color_schemes = (
            ["VESTA", "Jmol", "accessible"]
            + site_prop_types.get("scalar", [])
            + site_prop_types.get("categorical", [])
        )

        self.allowed_radius_schemes = (
            "atomic",
            "specified_or_average_ionic",
            "covalent",
            "van_der_waals",
            "atomic_calculated",
            "uniform",
        )

        if color_scheme not in self.allowed_color_schemes:
            warnings.warn(
                f"Color scheme {color_scheme} not available, "
                f"falling back to {self.default_color_scheme}."
            )
            color_scheme = self.default_color_scheme

        # if color-coding by a scalar site property, determine minimum and
        # maximum values for color scheme, will default to be symmetric
        # about zero
        if color_scheme in site_prop_types.get("scalar", []) and not cmap_range:
            props = np.array(
                [
                    p
                    for p in site_collection.site_properties[color_scheme]
                    if p is not None
                ]
            )
            prop_max = max([abs(min(props)), max(props)])
            prop_min = -prop_max
            cmap_range = (prop_min, prop_max)

        el_colors = EL_COLORS.copy()
        el_colors.update(
            self.generate_accessible_color_scheme_on_the_fly(site_collection)
        )

        self.categorical_colors = self.generate_categorical_color_scheme_on_the_fly(
            site_collection, site_prop_types
        )

        self.el_colors = el_colors
        self.site_prop_types = site_prop_types
        self.site_collection = site_collection
        self.color_scheme = color_scheme
        self.radius_scheme = radius_scheme
        self.cmap = cmap
        self.cmap_range = cmap_range

    @staticmethod
    def generate_accessible_color_scheme_on_the_fly(
        site_collection: SiteCollection,
    ) -> Dict[str, Dict[str, Tuple[int, int, int]]]:
        """
        e.g. for a color scheme more appropriate for people with color blindness

        Args:
            site_collection: SiteCollection

        Returns: A dictionary in similar format to EL_COLORS

        """

        color_scheme = {}

        all_species = set(
            chain.from_iterable(
                comp.keys() for comp in site_collection.species_and_occu
            )
        )
        all_elements = sorted([sp.as_dict()["element"] for sp in all_species])

        # thanks to https://doi.org/10.1038/nmeth.1618
        palette = [
            (0, 0, 0),  # 0, black
            (230, 159, 0),  # 1, orange
            (86, 180, 233),  # 2, sky blue
            (0, 158, 115),  #  3, bluish green
            (240, 228, 66),  # 4, yellow
            (0, 114, 178),  # 5, blue
            (213, 94, 0),  # 6, vermillion
            (204, 121, 167),  # 7, reddish purple
            (255, 255, 255),  #  8, white
        ]

        # similar to CPK, mapping element to palette index
        preferred_colors = {
            "O": 6,
            "N": 2,
            "C": 0,
            "H": 8,
            "F": 3,
            "Cl": 3,
            "Fe": 1,
            "Br": 7,
            "I": 7,
            "P": 1,
            "S": 4,
        }

        if len(set(all_elements)) > len(palette):
            warnings.warn(
                "Too many distinct types of site to use an accessible color scheme, "
                "some sites will be given the default color."
            )

        preferred_elements_present = [
            el for el in all_elements if el in preferred_colors.keys()
        ]

        colors_assigned = []
        for el in preferred_elements_present:
            if preferred_colors[el] not in colors_assigned:
                color_scheme[el] = palette[preferred_colors[el]]
                colors_assigned.append(preferred_colors[el])

        remaining_elements = [
            el for el in all_elements if el not in color_scheme.keys()
        ]
        remaining_palette = [
            c for idx, c in enumerate(palette) if idx not in colors_assigned
        ]

        for el in remaining_elements:
            if remaining_palette:
                color_scheme[el] = remaining_palette.pop()

        return {"accessible": color_scheme}

    @staticmethod
    def generate_categorical_color_scheme_on_the_fly(
        site_collection: SiteCollection, site_prop_types
    ) -> Dict[str, Dict[str, Tuple[int, int, int]]]:
        """
        e.g. for Wykcoff

        Args:
            site_collection: SiteCollection

        Returns: A dictionary in similar format to EL_COLORS

        """

        color_scheme = {}

        palette = Set1_9.colors

        for site_prop_name in site_prop_types.get("categorical", []):

            props = np.array(site_collection.site_properties[site_prop_name])
            props[props == None] = "None"

            le = LabelEncoder()
            le.fit(props)
            transformed_props = le.transform(props)

            # if we have more categories than availiable colors,
            # arbitrarily group some categories together
            if len(set(props)) > len(palette):
                warnings.warn(
                    "Too many categories for a complete categorical color scheme."
                )
            transformed_props = [
                p if p < len(palette) else -1 for p in transformed_props
            ]

            colors = {name: palette[p] for name, p in zip(props, transformed_props)}

            color_scheme[site_prop_name] = colors

        return color_scheme

    def get_color(self, sp: Union[Specie, Element], site: Optional[Site] = None) -> str:
        """
        Get a color to render a specific species. Optionally, you can provide
        a site for context, since ...

        Args:
            sp: Specie or Element
            site: Site

        Returns: Color

        """

        # allow manual override by user
        if site and "display_color" in site.properties:
            color = site.properties["display_color"]
            # TODO: next two lines due to change in API, will be removed
            if isinstance(color, list) and isinstance(color[0], str):
                color = color[0]
            if isinstance(color, list):
                return html5_serialize_simple_color(color)
            else:
                return html5_serialize_simple_color(html5_parse_legacy_color(color))

        if self.color_scheme in ("VESTA", "Jmol", "accessible"):
            el = sp.as_dict()["element"]
            color = self.el_colors[self.color_scheme].get(
                el, self.el_colors["Extras"].get(el, self.default_color)
            )

        elif self.color_scheme in self.site_prop_types.get("scalar", []):

            if not site:
                raise ValueError(
                    "Requires a site for context to get the "
                    "appropriate site property."
                )

            prop = site.properties[self.color_scheme]

            if prop:

                cmap = get_cmap(self.cmap)

                # normalize in [0, 1] range, as expected by cmap
                prop_min = self.cmap_range[0]
                prop_max = self.cmap_range[1]
                prop_normed = (prop - prop_min) / (prop_max - prop_min)

                color = [int(c * 255) for c in cmap(prop_normed)[0:3]]

            else:

                # fallback if site prop is None
                color = self.default_color

        elif self.color_scheme in self.site_prop_types.get("categorical", []):

            if not site:
                raise ValueError(
                    "Requires a site for context to get the "
                    "appropriate site property."
                )

            prop = site.properties[self.color_scheme]

            color = self.categorical_colors[self.color_scheme].get(
                prop, self.default_color
            )

        else:

            raise ValueError(
                f"Unknown color for {sp} and color scheme {self.color_scheme}."
            )

        return html5_serialize_simple_color(color)

    def get_radius(
        self, sp: Union[Specie, Element], site: Optional[Site] = None
    ) -> float:

        # allow manual override by user
        if site and "display_radius" in site.properties:
            return site.properties["display_radius"]

        if self.radius_scheme not in self.allowed_radius_schemes:
            raise ValueError(
                f"Unknown radius scheme {self.radius_scheme}, "
                f"choose from: {self.allowed_radius_schemes}."
            )

        radius = None
        if self.radius_scheme == "uniform":
            radius = self.uniform_radius
        elif self.radius_scheme == "atomic":
            radius = sp.atomic_radius
        elif (
            self.radius_scheme == "specified_or_average_ionic"
            and isinstance(sp, Specie)
            and sp.oxi_state
        ):
            radius = sp.ionic_radius
        elif self.radius_scheme == "specified_or_average_ionic":
            radius = sp.average_ionic_radius
        elif self.radius_scheme == "covalent":
            el = str(getattr(sp, "element", sp))
            radius = CovalentRadius.radius[el]
        elif self.radius_scheme == "van_der_waals":
            radius = sp.van_der_waals_radius
        elif self.radius_scheme == "atomic_calculated":
            radius = sp.atomic_radius_calculated

        if not radius:
            warnings.warn(
                "Radius unknown for {} and strategy {}, "
                "setting to 0.5.".format(sp, self.radius_scheme)
            )
            radius = self.fallback_radius

        return radius

    @staticmethod
    def analyze_site_props(site_collection: SiteCollection) -> Dict[str, List[str]]:
        """
        Returns: A dictionary with keys "scalar", "matrix", "vector", "categorical"
        and values of a list of site property names corresponding to each type
        """
        # (implicitly assumes all site props for a given key are same type)
        site_prop_names = defaultdict(list)
        for name, props in site_collection.site_properties.items():
            if isinstance(props[0], float) or isinstance(props[0], int):
                site_prop_names["scalar"].append(name)
            elif isinstance(props[0], list) and len(props[0]) == 3:
                if isinstance(props[0][0], list) and len(props[0][0]) == 3:
                    site_prop_names["matrix"].append(name)
                else:
                    site_prop_names["vector"].append(name)
            elif isinstance(props[0], str):
                site_prop_names["categorical"].append(name)
        return dict(site_prop_names)

    @staticmethod
    def get_species_str(sp: Union[Specie, Element]) -> str:
        """
        Args:
            sp: Specie or Element

        Returns: string representation
        """
        # TODO: add roman numerals for oxidation state for ease of readability
        # and then move this to pymatgen string utils ...
        return unicodeify_species(str(sp))

    def get_legend(self) -> Dict[str, Any]:

        # decide what we want the labels to be
        if self.color_scheme in ("Jmol", "VESTA", "accessible"):
            label = lambda site, sp: self.get_species_str(sp)
        elif self.color_scheme in self.site_prop_types.get("scalar", {}):
            label = lambda site, sp: f"{site.properties[self.color_scheme]:.2f}"
        elif self.color_scheme in self.site_prop_types.get("categorical", {}):
            label = lambda site, sp: f"{site.properties[self.color_scheme]}"
        else:
            raise ValueError(f"Color scheme {self.color_scheme} not known.")

        legend = defaultdict(list)

        # first get all our colors for different species
        for site in self.site_collection:
            for sp, occu in site.species.items():
                legend[self.get_color(sp, site)].append(label(site, sp))

        legend = {k: ", ".join(sorted(list(set(v)))) for k, v in legend.items()}

        color_options = []
        for site_prop_type in ("scalar", "categorical"):
            if site_prop_type in self.site_prop_types:
                for prop in self.site_prop_types[site_prop_type]:
                    color_options.append(prop)

        return {
            "composition": self.site_collection.composition.as_dict(),
            "colors": legend,
            "available_color_schemes": color_options,
        }



from crystal_toolkit.renderables.lattice import Lattice
from crystal_toolkit.renderables.moleculegraph import MoleculeGraph
from crystal_toolkit.renderables.molecule import Molecule
from crystal_toolkit.renderables.site import Site
from crystal_toolkit.renderables.structuregraph import StructureGraph
from crystal_toolkit.renderables.structure import Structure
from crystal_toolkit.renderables.volumetric import VolumetricData
from crystal_toolkit.renderables.phasediagram import PhaseDiagram



import itertools
from multiprocessing import cpu_count

from dash import dcc
from dash import html
import plotly.express as px
from dash import callback_context
from dash.dependencies import Input, Output
from dash.exceptions import PreventUpdate
from dash_mp_components import GraphComponent
from dscribe.descriptors import SOAP
from dscribe.kernels import REMatchKernel
from pymatgen.ext.matproj import MPRester
from pymatgen.analysis.chemenv.coordination_environments.chemenv_strategies import (
    SimplestChemenvStrategy,
)
from pymatgen.analysis.chemenv.coordination_environments.coordination_geometries import (
    AllCoordinationGeometries,
)
from pymatgen.analysis.chemenv.coordination_environments.coordination_geometry_finder import (
    LocalGeometryFinder,
)
from pymatgen.analysis.chemenv.coordination_environments.structure_environments import (
    LightStructureEnvironments,
)
from pymatgen.analysis.graphs import MoleculeGraph
from pymatgen.analysis.graphs import StructureGraph
from pymatgen.analysis.local_env import cn_opt_params, LocalStructOrderParams
from pymatgen.core.structure import Molecule, Structure
from pymatgen.io.ase import AseAtomsAdaptor
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.util.string import unicodeify_species, unicodeify
from sklearn.preprocessing import normalize

from crystal_toolkit.components.structure import StructureMoleculeComponent
from crystal_toolkit.core.legend import Legend
from crystal_toolkit.core.panelcomponent import PanelComponent
from crystal_toolkit.helpers.layouts import (
    get_data_list,
    Columns,
    Column,
    get_tooltip,
    cite_me,
    Loading,
    H5,
    Label,
)


def _get_local_order_parameters(structure_graph, n):
    """
    A copy of the method in pymatgen.analysis.local_env which
    can operate on StructureGraph directly.

    Calculate those local structure order parameters for
    the given site whose ideal CN corresponds to the
    underlying motif (e.g., CN=4, then calculate the
    square planar, tetrahedral, see-saw-like,
    rectangular see-saw-like order paramters).
    Args:
        structure_graph: StructureGraph object
        n (int): site index.
    Returns (Dict[str, float]):
        A dict of order parameters (values) and the
        underlying motif type (keys; for example, tetrahedral).
    """
    # TODO: move me to pymatgen once stable

    # code from @nisse3000, moved here from graphs to avoid circular
    # import, also makes sense to have this as a general NN method
    cn = structure_graph.get_coordination_of_site(n)
    if cn in [int(k_cn) for k_cn in cn_opt_params.keys()]:
        names = [k for k in cn_opt_params[cn].keys()]
        types = []
        params = []
        for name in names:
            types.append(cn_opt_params[cn][name][0])
            tmp = (
                cn_opt_params[cn][name][1] if len(cn_opt_params[cn][name]) > 1 else None
            )
            params.append(tmp)
        lostops = LocalStructOrderParams(types, parameters=params)
        sites = [structure_graph.structure[n]] + [
            connected_site.site
            for connected_site in structure_graph.get_connected_sites(n)
        ]
        lostop_vals = lostops.get_order_parameters(
            sites, 0, indices_neighs=[i for i in range(1, cn + 1)]
        )
        d = {}
        for i, lostop in enumerate(lostop_vals):
            d[names[i]] = lostop
        return d
    else:
        return None


class LocalEnvironmentPanel(PanelComponent):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.create_store("graph")
        self.create_store(
            "display_options",
            initial_data={"color_scheme": "Jmol", "color_scale": None},
        )

    @property
    def title(self):
        return "Local Environments"

    @property
    def description(self):
        return "Analyze the local chemical environments in your crystal."

    @property
    def loading_text(self):
        return "Analyzing environments"

    def contents_layout(self) -> html.Div:

        algorithm_choices = self.get_choice_input(
            label="Analysis method",
            kwarg_label="algorithm",
            state={"algorithm": "chemenv"},
            options=[
                {"label": "ChemEnv", "value": "chemenv"},
                {"label": "LocalEnv", "value": "localenv"},
                {"label": "Bonding Graph", "value": "bondinggraph"},
                {"label": "SOAP", "value": "soap"},
            ],
            help_str="Choose an analysis method to examine the local chemical environment. "
            "Several methods exist and there is no guaranteed correct answer, so try multiple!",
        )

        analysis = html.Div(id=self.id("analysis"))

        return html.Div([algorithm_choices, html.Br(), analysis, html.Br()])

    @staticmethod
    def get_graph_data(graph, display_options):

        color_scheme = display_options.get("color_scheme", "Jmol")

        nodes = []
        edges = []

        struct_or_mol = StructureMoleculeComponent._get_struct_or_mol(graph)
        legend = Legend(struct_or_mol, color_scheme=color_scheme)

        for idx, node in enumerate(graph.graph.nodes()):

            # TODO: fix for disordered
            node_color = legend.get_color(
                struct_or_mol[node].species.elements[0], site=struct_or_mol[node]
            )

            nodes.append(
                {
                    "id": node,
                    "title": f"{struct_or_mol[node].species_string} site "
                    f"({graph.get_coordination_of_site(idx)} neighbors)",
                    "color": node_color,
                }
            )

        for u, v, d in graph.graph.edges(data=True):

            edge = {"from": u, "to": v, "arrows": ""}

            to_jimage = d.get("to_jimage", (0, 0, 0))

            # TODO: check these edge weights
            if isinstance(struct_or_mol, Structure):
                dist = struct_or_mol.get_distance(u, v, jimage=to_jimage)
            else:
                dist = struct_or_mol.get_distance(u, v)
            edge["length"] = 50 * dist

            if to_jimage != (0, 0, 0):
                edge["arrows"] = "to"
                label = f"{dist:.2f} Å to site at image vector {to_jimage}"
            else:
                label = f"{dist:.2f} Å between sites"

            if label:
                edge["title"] = label

            # if 'weight' in d:
            #   label += f" {d['weight']}"

            edges.append(edge)

        return {"nodes": nodes, "edges": edges}

    def generate_callbacks(self, app, cache):

        super().generate_callbacks(app, cache)

        @app.callback(
            Output(self.id("analysis"), "children"),
            [Input(self.get_kwarg_id("algorithm"), "value")],
        )
        def run_algorithm(algorithm):

            algorithm = self.reconstruct_kwarg_from_state(
                callback_context.inputs, "algorithm"
            )

            if algorithm == "chemenv":

                state = {"distance_cutoff": 1.4, "angle_cutoff": 0.3}

                description = (
                    "The ChemEnv algorithm is developed by David Waroquiers et al. to analyze "
                    'local chemical environments. In this interactive app, the "SimplestChemenvStrategy" '
                    'and "LightStructureEnvironments" are used. For more powerful analysis, please use '
                    "the *pymatgen* code directly. Note that this analysis determines its own bonds independent "
                    "of those shown in the main crystal visualizer."
                )

                distance_cutoff = self.get_numerical_input(
                    label="Distance cut-off",
                    kwarg_label="distance_cutoff",
                    state=state,
                    help_str="Defines search radius by considering any atom within a radius "
                    "of the minimum nearest neighbor distance multiplied by the distance "
                    "cut-off.",
                    shape=(),
                )
                angle_cutoff = self.get_numerical_input(
                    label="Angle cut-off",
                    kwarg_label="angle_cutoff",
                    state=state,
                    help_str="Defines a tolerance whereby a neighbor atom is excluded if the solid angle "
                    "circumscribed by its Voronoi face is smaller than the angle tolerance "
                    "multiplied by the largest solid angle present in the crystal.",
                    shape=(),
                )

                return html.Div(
                    [
                        dcc.Markdown(description),
                        html.Br(),
                        cite_me(
                            cite_text="How to cite ChemEnv",
                            doi="10.26434/chemrxiv.11294480.v1",
                        ),
                        html.Br(),
                        distance_cutoff,
                        angle_cutoff,
                        html.Br(),
                        Loading(id=self.id("chemenv_analysis")),
                    ]
                )

            elif algorithm == "localenv":

                description = (
                    "The LocalEnv algorithm is developed by Nils Zimmerman et al. whereby "
                    "an 'order parameter' is calculated that measures how well that "
                    "environment matches an ideal polyhedra. The order parameter "
                    "is a number from zero to one, with one being a perfect match."
                )

                return html.Div(
                    [
                        dcc.Markdown(description),
                        html.Br(),
                        cite_me(
                            cite_text="How to cite LocalEnv",
                            doi="10.3389/fmats.2017.00034",
                        ),
                        html.Br(),
                        Loading(id=self.id("localenv_analysis")),
                    ]
                )

            elif algorithm == "bondinggraph":

                description = (
                    "This is an alternative way to display the same bonds present in the "
                    "visualizer. Here, the bonding is displayed as a crystal graph, with "
                    "nodes as atoms and edges as bonds. The graph visualization is shown in an "
                    "abstract two-dimensional space."
                )

                return html.Div(
                    [
                        dcc.Markdown(description),
                        html.Br(),
                        Loading(id=self.id("bondinggraph_analysis")),
                    ]
                )

            elif algorithm == "soap":

                state = {
                    "rcut": 5.0,
                    "nmax": 2,
                    "lmax": 2,
                    "sigma": 0.2,
                    "crossover": True,
                    "average": False,
                    "rbf": "gto",
                    "alpha": 0.1,
                    "threshold": 1e-4,
                    "metric": "linear",
                    "normalize_kernel": True,
                }

                description = (
                    'The "Smooth Overlap of Atomic Positions" (SOAP) descriptor provides information on the local '
                    "atomic environment by encoding that environment as a power spectrum derived from the "
                    "spherical harmonics of atom-centered gaussian densities. The SOAP formalism is complex but is "
                    "described well in [Bartók et al.](https://doi.org/10.1103/PhysRevB.87.184115) "
                    "and the REMatch similarity kernel in [De et al.](https://doi.org/10.1039/c6cp00415f) "
                    "The implementation of SOAP in this "
                    "web app is provided by [DScribe](https://doi.org/10.1016/j.cpc.2019.106949).  "
                    ""
                    "SOAP kernels are commonly used in machine learning applications. This interface is provided to "
                    "help gain intuition and exploration of the behavior of SOAP kernels."
                )

                rcut = self.get_numerical_input(
                    label="Radial cut-off /Å",
                    kwarg_label="rcut",
                    state=state,
                    help_str="The radial cut-off that defines the local region being considered",
                    shape=(),
                    min=1.0001,
                )

                nmax = self.get_numerical_input(
                    label="N max.",
                    kwarg_label="nmax",
                    state=state,
                    help_str="Number of radial basis functions",
                    shape=(),
                    is_int=True,
                    min=1,
                    max=9,
                )

                lmax = self.get_numerical_input(
                    label="L max.",
                    kwarg_label="lmax",
                    state=state,
                    help_str="Maximum degree of spherical harmonics",
                    shape=(),
                    is_int=True,
                    min=1,
                    max=9,
                )

                sigma = self.get_numerical_input(
                    label="Sigma",
                    kwarg_label="sigma",
                    state=state,
                    help_str="The standard deviation of gaussians used to build atomic density",
                    shape=(),
                    min=0.00001,
                )

                rbf = self.get_choice_input(
                    label="Radial basis function",
                    kwarg_label="rbf",
                    state=state,
                    help_str="Polynomial basis is faster, spherical gaussian based was used in original formulation",
                    options=[
                        {"label": "Spherical gaussian basis", "value": "gto"},
                        {"label": "Polynomial basis", "value": "polynomial"},
                    ],
                    style={"width": "16rem"},  # TODO: remove in-line style
                )

                crossover = self.get_bool_input(
                    label="Crossover",
                    kwarg_label="crossover",
                    state=state,
                    help_str="If enabled, the power spectrum will include all combinations of elements present.",
                )

                average = self.get_bool_input(
                    label="Average",
                    kwarg_label="average",
                    state=state,
                    help_str="If enabled, the SOAP vector will be averaged across all sites.",
                )

                alpha = self.get_numerical_input(
                    label="Alpha",
                    kwarg_label="alpha",
                    state=state,
                    help_str="Determines the entropic penalty in the REMatch kernel. As alpha goes to infinity, the "
                    "behavior of the REMatch kernel matches the behavior of the kernel where SOAP vectors "
                    "are averaged across all sites. As alpha goes to zero, the kernel matches the best match "
                    "kernel.",
                    shape=(),
                    min=0.00001,
                )

                threshold = self.get_numerical_input(
                    label="Sinkhorn threshold",
                    kwarg_label="threshold",
                    state=state,
                    help_str="Convergence threshold for the Sinkhorn algorithm. If values are too small, convergence "
                    "may not be possible, and calculation time will increase.",
                    shape=(),
                )

                metric = self.get_choice_input(
                    label="Metric",
                    kwarg_label="metric",
                    state=state,
                    help_str='See scikit-learn\'s documentation on "Pairwise metrics, Affinities and Kernels" '
                    "for an explanation of available metrics.",
                    options=[
                        # {"label": "Additive χ2", "value": "additive_chi2"},  # these seem to be unstable
                        # {"label": "Exponential χ2", "value": "chi2"},
                        {"label": "Linear", "value": "linear"},
                        {"label": "Polynomial", "value": "polynomial"},
                        {"label": "Radial basis function", "value": "rbf"},
                        {"label": "Laplacian", "value": "laplacian"},
                        {"label": "Sigmoid", "value": "sigmoid"},
                        {"label": "Cosine", "value": "cosine"},
                    ],
                    style={"width": "16rem"},  # TODO: remove in-line style
                )

                normalize_kernel = self.get_bool_input(
                    label="Normalize",
                    kwarg_label="normalize_kernel",
                    state=state,
                    help_str="Whether or not to normalize the resulting similarity kernel.",
                )

                # metric_kwargs = self.get_dict_input()

                return html.Div(
                    [
                        dcc.Markdown(description),
                        html.Br(),
                        H5("SOAP parameters"),
                        rcut,
                        nmax,
                        lmax,
                        sigma,
                        rbf,
                        crossover,
                        average,
                        html.Br(),  # TODO: remove all html.Br(), add appropriate styles instead
                        html.Br(),
                        html.Div(id=self.id("soap_analysis")),
                        html.Br(),
                        html.Br(),
                        H5("Similarity metric parameters"),
                        html.Div(
                            "This will calculate structural similarity scores from materials in the "
                            "Materials Project in the same chemical system. Note that for large chemical "
                            "systems this step can take several minutes."
                        ),
                        html.Br(),
                        alpha,
                        threshold,
                        metric,
                        # normalize_kernel,
                        html.Br(),
                        html.Br(),
                        Loading(id=self.id("soap_similarities")),
                    ]
                )

        def _get_soap_graph(feature, label):

            spectrum = {
                "data": [
                    {
                        "coloraxis": "coloraxis",
                        #'hovertemplate': 'x: %{x}<br>y: %{y}<br>color: %{z}<extra></extra>',
                        "type": "heatmap",
                        "z": feature.tolist(),
                    }
                ]
            }

            spectrum["layout"] = {
                "xaxis": {"visible": False},
                "yaxis": {"visible": False},
                "paper_bgcolor": "rgba(0,0,0,0)",
                "plot_bgcolor": "rgba(0,0,0,0)",
                "coloraxis": {
                    "colorscale": [
                        [0.0, "#0d0887"],
                        [0.1111111111111111, "#46039f"],
                        [0.2222222222222222, "#7201a8"],
                        [0.3333333333333333, "#9c179e"],
                        [0.4444444444444444, "#bd3786"],
                        [0.5555555555555556, "#d8576b"],
                        [0.6666666666666666, "#ed7953"],
                        [0.7777777777777778, "#fb9f3a"],
                        [0.8888888888888888, "#fdca26"],
                        [1.0, "#f0f921"],
                    ],
                    "showscale": False,
                },
                "margin": {"l": 0, "b": 0, "t": 0, "r": 0, "pad": 0},
                # "height": 20*feature.shape[0],  # for fixed size plots
                # "width": 20*feature.shape[1]
            }

            return Columns(
                [
                    Column(Label(label), size="1"),
                    Column(
                        dcc.Graph(
                            figure=spectrum,
                            config={"displayModeBar": False},
                            responsive=True,
                            style={"height": "60px"},
                        )
                    ),
                ]
            )

        @app.callback(
            Output(self.id("soap_analysis"), "children"),
            [Input(self.id(), "data"), Input(self.get_all_kwargs_id(), "value")],
        )
        def update_soap_analysis(struct, all_kwargs):

            if not struct:
                raise PreventUpdate

            struct = self.from_data(struct)
            kwargs = self.reconstruct_kwargs_from_state(callback_context.inputs)

            # TODO: make sure is_int kwarg information is enforced so that int() conversion is unnecessary
            desc = SOAP(
                species=[e.number for e in struct.composition.elements],
                sigma=kwargs["sigma"],
                rcut=kwargs["rcut"],
                nmax=int(kwargs["nmax"]),
                lmax=int(kwargs["lmax"]),
                periodic=True,
                crossover=kwargs["crossover"],
                sparse=False,
                average=kwargs["average"],
            )

            adaptor = AseAtomsAdaptor()
            atoms = adaptor.get_atoms(struct)
            feature = normalize(desc.create(atoms, n_jobs=cpu_count()))

            return _get_soap_graph(feature, "SOAP vector for this material")

        @cache.memoize(timeout=360)
        def _get_all_structs_from_elements(elements):
            structs = {}
            all_chemsyses = []
            for i in range(len(elements)):
                for els in itertools.combinations(elements, i + 1):
                    all_chemsyses.append("-".join(sorted(els)))

            with MPRester() as mpr:
                docs = mpr.query(
                    {"chemsys": {"$in": all_chemsyses}}, ["task_id", "structure"],
                )
            structs.update({d["task_id"]: d["structure"] for d in docs})
            return structs

        @app.callback(
            Output(self.id("soap_similarities"), "children"),
            [Input(self.id(), "data"), Input(self.get_all_kwargs_id(), "value")],
        )
        def update_soap_similarities(struct, all_kwargs):

            if not struct:
                raise PreventUpdate

            structs = {"input": self.from_data(struct)}
            kwargs = self.reconstruct_kwargs_from_state(callback_context.inputs)

            elements = [str(el) for el in structs["input"].composition.elements]
            structs.update(_get_all_structs_from_elements(elements))

            if not structs:
                raise PreventUpdate

            elements = {
                elem for s in structs.values() for elem in s.composition.elements
            }
            # TODO: make sure is_int kwarg information is enforced so that int() conversion is unnecessary
            desc = SOAP(
                species=[e.number for e in elements],
                sigma=kwargs["sigma"],
                rcut=kwargs["rcut"],
                nmax=int(kwargs["nmax"]),
                lmax=int(kwargs["lmax"]),
                periodic=True,
                crossover=kwargs["crossover"],
                sparse=False,
                average=kwargs["average"],
            )

            adaptor = AseAtomsAdaptor()
            atomss = {
                mpid: adaptor.get_atoms(struct) for mpid, struct in structs.items()
            }

            print(f"Calculating {len(atomss)} SOAP vectors")
            features = {
                mpid: normalize(desc.create(atoms, n_jobs=cpu_count()))
                for mpid, atoms in atomss.items()
            }

            re = REMatchKernel(
                metric=kwargs["metric"],
                alpha=kwargs["alpha"],
                threshold=kwargs["threshold"],
                # normalize_kernel=kwargs["normalize_kernel"],
            )

            print("Calculating similarity kernel")
            similarities = {
                mpid: re.get_global_similarity(
                    re.get_pairwise_matrix(features["input"], feature)
                )
                for mpid, feature in features.items()
                if mpid != "input"
            }

            sorted_mpids = sorted(similarities.keys(), key=lambda x: -similarities[x])

            print("Generating similarity graphs")
            # TODO: was much slower using px.imshow (see prev commit)
            all_graphs = [
                _get_soap_graph(
                    features[mpid],
                    [
                        html.Span(
                            f"{unicodeify(structs[mpid].composition.reduced_formula)}"
                        ),
                        dcc.Markdown(f"[{mpid}](https://materialsproject.org/{mpid})"),
                        html.Span(f"{similarities[mpid]:.5f}"),
                    ],
                )
                for mpid in sorted_mpids
            ]

            print("Returning similarity graphs")
            return html.Div(all_graphs)

        @app.callback(
            Output(self.id("localenv_analysis"), "children"),
            [Input(self.id("graph"), "data")],
        )
        def update_localenv_analysis(graph):

            if not graph:
                raise PreventUpdate

            graph = self.from_data(graph)

            return html.Div([str(_get_local_order_parameters(graph, 0))])

        @app.callback(
            Output(self.id("bondinggraph_analysis"), "children"),
            [
                Input(self.id("graph"), "data"),
                Input(self.id("display_options"), "data"),
            ],
        )
        def update_bondinggraph_analysis(graph, display_options):

            if not graph:
                raise PreventUpdate

            graph = self.from_data(graph)
            display_options = self.from_data(display_options)

            graph_data = self.get_graph_data(graph, display_options)

            options = {
                "interaction": {
                    "hover": True,
                    "tooltipDelay": 0,
                    "zoomView": False,
                    "dragView": False,
                },
                "edges": {
                    "smooth": {"type": "dynamic"},
                    "length": 250,
                    "color": {"inherit": "both"},
                },
                "physics": {
                    "solver": "forceAtlas2Based",
                    "forceAtlas2Based": {"avoidOverlap": 1.0},
                    "stabilization": {"fit": True},
                },
            }

            return html.Div(
                [GraphComponent(graph=graph_data, options=options)],
                style={"width": "65vmin", "height": "65vmin"},
            )

        @app.callback(
            Output(self.id("chemenv_analysis"), "children"),
            [
                Input(self.id(), "data"),
                Input(self.get_kwarg_id("distance_cutoff"), "value"),
                Input(self.get_kwarg_id("angle_cutoff"), "value"),
            ],
        )
        def get_chemenv_analysis(struct, distance_cutoff, angle_cutoff):

            if not struct:
                raise PreventUpdate

            struct = self.from_data(struct)
            kwargs = self.reconstruct_kwargs_from_state(callback_context.inputs)
            distance_cutoff = kwargs["distance_cutoff"]
            angle_cutoff = kwargs["angle_cutoff"]

            # TODO: remove these brittle guard statements, figure out more robust way to handle multiple input types
            if isinstance(struct, StructureGraph):
                struct = struct.structure

            def get_valences(struct):
                valences = [getattr(site.specie, "oxi_state", None) for site in struct]
                valences = [v for v in valences if v is not None]
                if len(valences) == len(struct):
                    return valences
                else:
                    return "undefined"

            # decide which indices to present to user
            sga = SpacegroupAnalyzer(struct)
            symm_struct = sga.get_symmetrized_structure()
            inequivalent_indices = [
                indices[0] for indices in symm_struct.equivalent_indices
            ]
            wyckoffs = symm_struct.wyckoff_symbols

            lgf = LocalGeometryFinder()
            lgf.setup_structure(structure=struct)

            se = lgf.compute_structure_environments(
                maximum_distance_factor=distance_cutoff + 0.01,
                only_indices=inequivalent_indices,
                valences=get_valences(struct),
            )
            strategy = SimplestChemenvStrategy(
                distance_cutoff=distance_cutoff, angle_cutoff=angle_cutoff
            )
            lse = LightStructureEnvironments.from_structure_environments(
                strategy=strategy, structure_environments=se
            )
            all_ce = AllCoordinationGeometries()

            envs = []
            unknown_sites = []

            for index, wyckoff in zip(inequivalent_indices, wyckoffs):

                datalist = {
                    "Site": unicodeify_species(struct[index].species_string),
                    "Wyckoff Label": wyckoff,
                }

                if not lse.neighbors_sets[index]:
                    unknown_sites.append(f"{struct[index].species_string} ({wyckoff})")
                    continue

                # represent the local environment as a molecule
                mol = Molecule.from_sites(
                    [struct[index]] + lse.neighbors_sets[index][0].neighb_sites
                )
                mol = mol.get_centered_molecule()
                mg = MoleculeGraph.with_empty_graph(molecule=mol)
                for i in range(1, len(mol)):
                    mg.add_edge(0, i)

                view = html.Div(
                    [
                        StructureMoleculeComponent(
                            struct_or_mol=mg,
                            disable_callbacks=True,
                            id=f"{struct.composition.reduced_formula}_site_{index}",
                            scene_settings={"enableZoom": False, "defaultZoom": 0.6},
                        )._sub_layouts["struct"]
                    ],
                    style={"width": "300px", "height": "300px"},
                )

                env = lse.coordination_environments[index]
                co = all_ce.get_geometry_from_mp_symbol(env[0]["ce_symbol"])
                name = co.name
                if co.alternative_names:
                    name += f" (also known as {', '.join(co.alternative_names)})"

                datalist.update(
                    {
                        "Environment": name,
                        "IUPAC Symbol": co.IUPAC_symbol_str,
                        get_tooltip(
                            "CSM",
                            "The continuous symmetry measure (CSM) describes the similarity to an "
                            "ideal coordination environment. It can be understood as a 'distance' to "
                            "a shape and ranges from 0 to 100 in which 0 corresponds to a "
                            "coordination environment that is exactly identical to the ideal one. A "
                            "CSM larger than 5.0 already indicates a relatively strong distortion of "
                            "the investigated coordination environment.",
                        ): f"{env[0]['csm']:.2f}",
                        "Interactive View": view,
                    }
                )

                envs.append(get_data_list(datalist))

            # TODO: switch to tiles?
            envs_grouped = [envs[i : i + 2] for i in range(0, len(envs), 2)]
            analysis_contents = []
            for env_group in envs_grouped:
                analysis_contents.append(
                    Columns([Column(e, size=6) for e in env_group])
                )

            if unknown_sites:
                unknown_sites = html.Strong(
                    f"The following sites were not identified: {', '.join(unknown_sites)}. "
                    f"Please try changing the distance or angle cut-offs to identify these sites, "
                    f"or try an alternative algorithm such as LocalEnv."
                )
            else:
                unknown_sites = html.Span()

            return html.Div([html.Div(analysis_contents), html.Br(), unknown_sites])
